<?php
/**
 * Matomo - free/libre analytics platform
 *
 * @link https://matomo.org
 * @license http://www.gnu.org/licenses/gpl-3.0.html GPL v3 or later
 */

namespace Matomo\Ini;

/**
 * Exception when writing a INI configuration.
 */
class IniWritingException extends \Exception
{
}
